// Copyright 2010, 2011, 2012 by Jon Dart. All Rights Reserved.
#include "board.h"
#include "notation.h"
#include "legal.h"
#include "hash.h"
#include "globals.h"
#include "chessio.h"
#include "util.h"
#include "search.h"
#include <iostream>
#include <fstream>
#include <ctype.h>
#include <map>
extern "C" {
#include <math.h>
};

static const int MIN_PLY = 16;

int CDECL main(int argc, char **argv)
{
    Bitboard::init();
    initOptions(argv[0]);
    Attacks::init();
    Scoring::init();
    options.search.hash_table_size = 64*1024*1024;
    if (!initGlobals(argv[0], false)) {
        cleanupGlobals();
        exit(-1);
    }
    atexit(cleanupGlobals);
    delayedInit();
    if (EGTBMenCount) {
        cerr << "Initialized tablebases" << endl;
    }
    options.book.book_enabled = options.log_enabled = 0;

    int arg = 1;
    unsigned minMoves = 30;
    int minELO = 0;
    // time limit in ms.
    int timeLimit = 100;

    if (argc ==1) {
        cerr << "usage: select.exe -t <search time> <input file>" << endl;
        return -1;
    }
    else {
        for (;arg < argc;++arg) {
            if (argv[arg][0] == '-') {
                if (argv[arg][1] == 't') {
                    ++arg;
                    if (arg < argc) {
                        timeLimit = atoi(argv[arg]);
                    }
                }
            }
            else
                break;
        }
        SearchController *searcher = new SearchController();
        Statistics stats;
        for (;arg < argc;arg++) {
            ifstream pgn_file( argv[arg], ios::in);
            int c;
            int count = 0;
            hash_t game_hash;
            ColorType side;
            string result, white, black;
            if (!pgn_file.good()) {
                cerr << "could not open file " << argv[arg] << endl;
                exit(-1);
            }
            else {
                ArasanVector<ChessIO::Header> hdrs;

                Board board;
                while (!pgn_file.eof()) {

                    long first;
                    MoveArray moves;
                    board.reset();
                    side = White;
                    while (pgn_file.good() && (c = pgn_file.get()) != EOF) {
                        if (c=='[') {
                            pgn_file.putback(c);
                            break;
                        }
                    }
                    hdrs.removeAll();
                    ChessIO::collect_headers(pgn_file,hdrs,first);

                    game_hash = 0;
                    int mate = 0;
                    int valid = 1;
                    int have_scores =  0;
                    int var = 0;
                    int done = 0;
                    string num;
                    int halfmoves = 0;
                    float score;
                    while (!done) {
                        ChessIO::Token tok = ChessIO::get_next_token(pgn_file);
                        switch(tok.type) {
                        case ChessIO::Eof:
                            done = true;
                            break;
                        case ChessIO::Number:
                            if (var) continue;
                            num = tok.val.c_str();
                            break;
                        case ChessIO::GameMove: {
                            if (var) continue;  
                            // parse the move
                            Move m;
                            m = Notation::value(board,board.sideToMove(),
                                                Notation::SAN_IN,tok.val);
                            if (IsNull(m)/* ||
                                            !legalMove(board,StartSquare(m), DestSquare(m))*/) {
                                // echo to both stdout and stderr
                                cerr << "Illegal move: " << tok.val << endl;
                                cout << "Illegal move: " << tok.val << endl;
                                valid = 0;

                            }
                            else if (valid) {
                                string result;
                                Notation::image(board,m,Notation::SAN_OUT,result);
                                BoardState bs = board.state;
                                moves.add_move(board,bs,m,result,false);

                                board.doMove(m);
                                ++halfmoves;
                            }
                            side = OppositeColor(side);
                            break;
                        }
                        case ChessIO::Unknown:
                            if (strcmp(tok.val.c_str(),"(") == 0)
                                ++var;
                            else if (strcmp(tok.val.c_str(),")")==0) {
                                --var;   
                            } else {
                                cerr << "Unrecognized text: " << tok.val << endl;
                                valid = 0;
                            }
                            break;
                        case ChessIO::Comment:
                            break;
                        case ChessIO::Result: {
                            result = tok.val;
                        }
                        ++done;
                        break;
                        } // end switch
                    }
                    // done with a game
                    if (moves.num_moves() < minMoves || !valid) {
                        continue;
                    }
                    int choice =  MIN_PLY;
                    Board b;
                    float last_score = (float)Scoring::INVALID_SCORE;
                    for (int i = 0; i < Util::Min(250,moves.num_moves()); i++) {
                        b.doMove(moves.move(i));
                        if (i == choice) {
                            /*
                              searcher->findBestMove(b,
                              FixedDepth,
                              999999,
                              0, //extra time
                              6,  // ply limit
                              false,  // background
                              false, // UCI
                              stats,
                              Silent);
                            score = (float)stats.value/100.0F;
                            if (score >= -5.0F && score <= 5.0F) {
                                cout << b << " " << result << endl;
                                last_score = score;
                            }
                            if (i >= 150 && score == 0.0F &&
                                last_score == 0.0F) break;
                            choice += 11;
                            */
                            cout << b << " " << result << endl;
                            choice += 21;
                        }
                    }
                }
                pgn_file.close();
            }
        }
    }

    return 0;
}
